package org.sinev.igor.ps.elementtype;

import java.util.EnumMap;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Maintains constants for ElementType enumeration
 * 
 * @author Igor Sinev
  */
final class Constants {
	
	private static final String ourRealPatternPart = "[+\\-]?(((0|[1-9])\\d*(\\.\\d*)?)|(\\.\\d*))";
	private static final String ourOrdinaryCharacter = "[^()<>\\[\\]{}/%\\n\\r\\f \\t\\00]";
	
	/**
	 * Stores patterns for different element types.
	 */
	static HashMap<String, String> ourPatterns = new HashMap<String, String>();
	static {
		ourPatterns.put("BOOL", "(true)|(false)");
		ourPatterns.put("INT", "[+\\-]?(0|[1-9]\\d*)");
		ourPatterns.put("REAL", ourRealPatternPart + "([eE]" + ourRealPatternPart + ")?");
		ourPatterns.put("RADIX", "([1-9]|(1\\d)|(2[0-6]))#(\\d|[a-fA-F])+");		
		ourPatterns.put("CHARACTER_STRING", "\\(.*\\)");
		ourPatterns.put("LITERAL_NAME", "/" + ourOrdinaryCharacter + "*");
		ourPatterns.put("PROCEDURE", "\\{.*\\}");		
		ourPatterns.put("ARRAY", "\\[.*\\]");
		ourPatterns.put("NAME", ourOrdinaryCharacter + "*");
		ourPatterns.put("WHITESPACE", "[\\n\\r\\f \\t\\00]*");
		ourPatterns.put("COMMENT", "%.*");
		ourPatterns.put("HEXADECIMAL_STRING", "<.*>"); 
	}
		
}

/**
 * Enumeration that specifies type of structure elements.
 * 
 * @author Igor Sinev
 *
 */
public enum ElementType {
	WHITESPACE,
	
	BOOL,
	INT,
	REAL,
	RADIX,
	CHARACTER_STRING,
	HEXADECIMAL_STRING,
	LITERAL_NAME,
	ARRAY,
	PROCEDURE,
	NAME,
	COMMENT,
	
	NUMBER,
	STRING,
	ANY_TYPE,
			
	UNDEFINED;			
	
	private Pattern myPattern;	
	private ElementType() {		
		String s = Constants.ourPatterns.get(this.name());
		if (s == null)
			s = String.valueOf(0);
		myPattern = Pattern.compile(s, Pattern.DOTALL);
		//System.out.println(this + " " + pattern);
	}
	
	/**
	 * Reads the element type from string.
	 * 
	 * @param s name of the type.
	 * 
	 * @return read type.
	 */
	public static ElementType getByName(String s) {
		for (ElementType e : ElementType.values())
			if (e.toString().equals(s))
				return e;
		return ElementType.UNDEFINED;
	}
	
	
	/**
	 * Determines the type of the given element.
	 * 
	 * @param s element string representation
	 * 
	 * @return type of specified element
	 */
	public static ElementType getFromString(String s) {
		Matcher m;
		for (ElementType e : ElementType.values()) {
			//System.out.println(e);
			//if (s == null)
			//	System.out.println("!!!");
			m = e.myPattern.matcher(s);
			if (m.matches())
				return e;
		}
		
		return UNDEFINED;		
	}
	
	/**
	 * Defines types compability.
	 */
	static EnumMap<ElementType, EnumSet<ElementType>> ourMatchesMap = null;
	
	private static void createMatchesMap() { 
		ourMatchesMap = new EnumMap<ElementType, EnumSet<ElementType>>(ElementType.class);
		for (ElementType e : ElementType.values()) {
			ourMatchesMap.put(e, EnumSet.of(ANY_TYPE, e));
		}
		for (ElementType e : ElementType.values())
			ourMatchesMap.get(ANY_TYPE).add(e);
		ourMatchesMap.get(INT).add(NUMBER);
		ourMatchesMap.get(REAL).add(NUMBER);
		ourMatchesMap.get(CHARACTER_STRING).add(STRING);
		ourMatchesMap.get(HEXADECIMAL_STRING).add(STRING);		
	}	

	/**
	 * Determines whether given type can be 
	 * an instance of specified type.
	 * 
	 * @param t type that is supposed to be 
	 * generalization of the given. 
	 * 
	 * @return true if specified type is 
	 * generalization of given. 
	 */
	public boolean matches(ElementType t) {
		if (ourMatchesMap == null)
			createMatchesMap();
		return ourMatchesMap.get(this).contains(t);
	}
		
	/**
	 * Finds a type that is generalization of both current 
	 * type and specified.
	 * 
	 * @param t type to find common with.
	 * 
	 * @return generalization of two types.
	 */
	public ElementType combinateWith(ElementType t) {
		if (t.matches(this))
			return this;
		else if (this.matches(t))
			return t;
		else
			return ElementType.ANY_TYPE;
	}
}