package org.sinev.igor.ps.model.impl;

import java.util.HashMap;

import org.sinev.igor.ps.PSParseException;
import org.sinev.igor.ps.model.IDocument;
import org.sinev.igor.ps.model.IElement;
import org.sinev.igor.ps.model.IPSEArrayElement;
import org.sinev.igor.ps.model.IPSEBoolean;
import org.sinev.igor.ps.model.IPSECharacterString;
import org.sinev.igor.ps.model.IPSEComment;
import org.sinev.igor.ps.model.IPSEHexadecimalString;
import org.sinev.igor.ps.model.IPSEInteger;
import org.sinev.igor.ps.model.IPSEOperator;
import org.sinev.igor.ps.model.IPSEProcedureElement;
import org.sinev.igor.ps.model.IPSEReal;
import org.sinev.igor.ps.model.IPSEUndefinedElement;
import org.sinev.igor.ps.model.IPSEWhitespace;

/**
 * Utility class that creates new elements.
 * 
 * @author Igor Sinev 
 */
public class ElementFactory {
	/**
	 * Creates a boolean element from it's string representation.
	 * 
	 * @param representation element representation.
	 * 
	 * @param d document this element is created for
	 * 
	 * @return created element.
	 */
	public static IPSEBoolean createBooleanElement(String representation, IDocument d) {
		return new PSEBoolean(representation, d);
	}
	
	/**
	 * Creates an integer element from it's string representation.
	 * 
	 * @param representation element representation.
	 * 
	 * @param d document this element is created for
	 * 
	 * 
	 * @return created element.
	 */
	public static IPSEInteger createIntegerElement(String representation, IDocument d) {
		return new PSEInteger(representation, d);
	}
	
	/**
	 * Creates an integer element from it's value.
	 * 
	 * @param representation element value.
	 * 
	 * @param d document this element is created for
	 * 
	 * 
	 * @return created element.
	 */
	public static IPSEInteger createIntegerElement(int v, IDocument d) {
		return new PSEInteger(v, d);
	}
	
	/**
	 * Creates a real element from it's string representation.
	 * 
	 * @param representation element representation.
	 * 
	 * @param d document this element is created for
	 * 
	 * 
	 * @return created element.
	 */
	public static IPSEReal createRealElement(String representation, IDocument d) {
		return new PSEReal(representation, d);
	}
	
	/**
	 * Creates an real element from it's string representation.
	 * 
	 * @param representation element value.
	 * 
	 * @param d document this element is created for
	 * 
	 * 
	 * @return created element.
	 */
	public static IPSEReal createRealElement(double v, IDocument d) {
		return new PSEReal(v, d);
	}

	/**
	 * Creates a character string element from it's string representation.
	 * 
	 * @param representation element representation.
	 * 
	 * @param v contents of the string.
	 * 
	 * @param d document this element is created for
	 * 
	 * 
	 * @return created element.
	 */
	public static IPSECharacterString createCharacterStringElement(String representation, String v, IDocument d) {
		return new PSECharacterString(representation, v, d);
	}
	
	/**
	 * Creates a hexadecimal string element from it's string representation.
	 * 
	 * @param representation element representation.
	 * 
	 * @param v contents of the string.
	 * 
	 * @param d document this element is created for
	 * 
	 * @return created element.
	 */
	public static IPSEHexadecimalString createHexadecimalStringElement(String representation, byte[] v, IDocument d) {
		return new PSEHexadecimalString(representation, v, d);
	}

	/**
	 * Creates an operator element from it's string representation.
	 * 
	 * @param representation element representation.
	 * 
	 * @param d document this element is created for
	 * 
	 * @return created element.
	 */		
	public static IPSEOperator createOperatorElement(String representation, IDocument d) {
		return new PSEOperator(representation, d);
	}

	private static HashMap<String, IPSEWhitespace> ourWhitespaces = new HashMap<String, IPSEWhitespace>(); 
	/**
	 * Creates a whitespace element from it's string representation.
	 * 
	 * @param representation element representation.
	 * 
	 * @param d document this element is created for
	 * 
	 * @return created element.
	 */
	public static IPSEWhitespace createWhitespaceElement(String representation, IDocument d) {
		IPSEWhitespace w = ourWhitespaces.get(representation); 
		if (w == null) {
			w = new PSEWhitespace(representation, d);
			ourWhitespaces.put(representation, w);
		}
		return w;		
	}

	/**
	 * Creates a comment element.
	 * 
	 * @param representation comment text
	 * 
	 * @param d document this element is created for
	 * 
	 * @return created element.
	 */
	public static IPSEComment createCommentElement(String representation, IDocument d) {
		return new PSEComment(representation, d);
	}
	
	/**
	 * Forms a string from the contents of IElement array.
	 * 
	 * @param elements given array
	 * 
	 * @return string representation of this array
	 */
	private static String elementArrayToString(IElement[] elements) {
		StringBuffer res = new StringBuffer();
		for (IElement e : elements)
			res.append(e.getRepresentation());
		return res.toString();
	}
	
	/**
	 * Creates an element of undefined type.
	 * 
	 * @param representation element representation.
	 * 
	 * @param d document this element is created for
	 * 
	 * @return created element.
	 */
	public static IPSEUndefinedElement createUndefinedElement(String representation, IDocument d) {
		return new PSEUndefinedElement(representation, d);
	}

	/**
	 * Creates an array element.
	 * 
	 * @param contents array elements.
	 * 
	 * @param d document this element is created for
	 * 
	 * @return created array element.
	 */
	public static IPSEArrayElement createArrayElement(IElement[] contents, IDocument d) {
		return new PSEArrayElement(contents, 
				new StringBuffer("[").append(elementArrayToString(contents))
				.append(']').toString(),
				d
			);
	}

	/**
	 * Creates a procedure element.
	 * 
	 * @param contents procedure elements.
	 * 
	 * @param d document this element is created for
	 * 
	 * @return created element.
	 * @throws PSParseException 
	 */
	public static IPSEProcedureElement createProcedureElement(IElement[] contents, IDocument d) throws PSParseException {
		return new PSEProcedureElement(contents, 
				new StringBuffer("{").append(elementArrayToString(contents))
				.append('}').toString(),
				d				
			);
	}
}
