package org.sinev.igor.ps.model.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.sinev.igor.ps.PSParseException;
import org.sinev.igor.ps.elementtype.ElementType;
import org.sinev.igor.ps.model.IDocument;
import org.sinev.igor.ps.model.IElement;
import org.sinev.igor.ps.model.IPSEOperator;
import org.sinev.igor.ps.model.IStackObject;
import org.sinev.igor.ps.operatordescription.OperatorDescription;

/**
 * Represents an operator element of some PS document.
 * 
 * @author Igor Sinev
 */
public class PSEOperator extends FunctionalElement implements IPSEOperator {		
	/**
	 * Syntax description of the operator.
	 */
	private OperatorDescription myDescription;
	
	PSEOperator(String representation, IDocument d) {
		super(representation, d);
		myDescription = d.getNamingMap().getDescription(representation);
		myArguments = new ArrayList<IStackObject>(myDescription.getArguments().size());
		myReturns = new ArrayList<IElement>(myDescription.getReturns().size());
	}
	
	public List<IStackObject> getArguments() {
		return Collections.unmodifiableList(myArguments);		
	}

	/**
	 * Gets PostScript syntax description bound to the operator.
	 * 
	 * @return operator description.	 
	 */
	public OperatorDescription getDescription() {
		return myDescription;
	}
	
	@Override
	void putToStack(ElementStructure.StorageStack s) throws PSParseException {
		int i = 0;
		ElementType a = null;
		for (ElementType t : myDescription.getArguments()) {			
			IStackObject o = s.pop(t);			
			ElementType ot = (o == null) ? t : o.getType();						
			if (!ot.matches(t))
				throw new PSParseException("Element type doesn't match: expected " + t + ", found " + ot + ".");
			
			// for arithmetic
			if (a == null)
				a = ot;
			else
				a = a.combinateWith(ot);
			
			myArguments.add(o);
			i++;
		}
		Collections.reverse(myArguments);
		
		for (i = 0; i < myDescription.getReturns().size(); i++) {
			if (myDescription.getGroup().indexOf("Arithmetic") > -1)
				s.push(new CalculatedStackObject(this, i, a));
			else
				s.push(new CalculatedStackObject(this, i));
		}
	}
		
	@Override
	public boolean isInterpretable() {
		return true;
	}
        
	@Override
	protected void structureDebugOutputWriteSubtree(StringBuffer target) {
		for (int i = 0; i < structureDebugOutputIndent; i++)
            target.append("  ");                     
        target.append(this).append(" (").append("Operator").append(")\n");
		for (IStackObject e : myArguments) {
        	structureDebugOutputIndent++;
        	DocumentDebugWriter.callStructureDebugOutput(e, target);            
            structureDebugOutputIndent--;
        }
	}
	
	@Override
	public ElementType getReturnType(int myNumber) {
		return getDescription().getReturns().get(myNumber);
	}

	@Override
	protected void calculate() {
		if (myCalculated)
			return;		
		ArrayList<IElement> args = new ArrayList<IElement>(myArguments.size());		
		for (IStackObject so : myArguments)
			args.add(so.getElement());
		myReturns = new ArrayList<IElement>(myDescription.calculate(args, (Document) myDocument));
		myCalculated = true;
	}	
}
