package org.sinev.igor.ps.reader;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Stack;

import org.sinev.igor.ps.PSParseException;
import org.sinev.igor.ps.model.IDocument;
import org.sinev.igor.ps.model.IElement;
import org.sinev.igor.ps.model.impl.ElementFactory;
import org.sinev.igor.ps.elementtype.ElementType;
import org.sinev.igor.ps.reader.TerminatedStringReader.TerminatorChecker;

/**
 * Provides methods to get single elements from the reader.
 * 
 * @author Igor 
 */
public class ElementReader {	
	
	/**
	 * Creates an element that would be put to a specified 
	 * document from it's string representation.
	 * 
	 * @param representation source string
	 * @param d target document
	 *  
	 * @return created element
	 * 
	 * @throws PSParseException
	 */
	private static IElement createElementFromString(String representation, IDocument d) throws PSParseException {
		switch (ElementType.getFromString(representation)) {
		case BOOL:
			return ElementFactory.createBooleanElement(representation, d);
		case INT:
			return ElementFactory.createIntegerElement(representation, d);
		case REAL:
			return ElementFactory.createRealElement(representation, d);
		case CHARACTER_STRING:
			return ElementFactory.createCharacterStringElement(representation, parseCharacterString(representation.substring(1, representation.length() - 1)), d);
		case HEXADECIMAL_STRING:
			return ElementFactory.createHexadecimalStringElement(representation, parseHexadecimalString(representation.substring(1, representation.length() - 1)), d);
		case WHITESPACE:
			return ElementFactory.createWhitespaceElement(representation, d);
		case NAME:
			return ElementFactory.createOperatorElement(representation, d);		
		case COMMENT:
			return ElementFactory.createCommentElement(representation, d);
		default:
			return ElementFactory.createUndefinedElement(representation, d);
		}		
	}
	
	private static byte[] parseHexadecimalString(String s) {
		byte[] res = new byte[(s.length() + 1) / 2];
		int i = 0;
		int ri = 0;
		while (i < s.length()) {
			res[ri] = 0;			
			for (int j = 0; i < s.length(); j++) {
				char c = s.charAt(i++);			
				byte b;
				if ('0' <= c && c <= '9')
					b = (byte) (c - '0');
				else
					b = (byte) (c - 'a');
				res[ri] = b;				
				if (j == 1)
					break;
				res[ri] <<= 4;				
			}
			ri++;
		}
		return res;
	}

	private static TerminatorChecker commentTerminatorChecker = new TerminatorChecker() {
			public boolean isTerminator(char c) {
				return Routines.isNewLine(c);
			}					
		};		
	private static TerminatorChecker whitespaceTerminatorChecker = new TerminatorChecker() {
		public boolean isTerminator(char c) {
			return !Routines.isWhitespace(c);
		}					
	};
	private static TerminatorChecker regularTerminatorChecker = new TerminatorChecker() {
		public boolean isTerminator(char c) {
			return Routines.isWhitespace(c) || c == '%' || (!arrayTerminatingCharacterStack.empty() && arrayTerminatingCharacterStack.peek().equals(c));
		}					
	};
				
	private static Stack<Character> arrayTerminatingCharacterStack = new Stack<Character>();
	/**
	 * Forms one element for a specified document from 
	 * a token read from reader.
	 *  
	 * @param r reader to operate with
	 * @param d document for this element.
	 * @return element that was read.
	 * @throws IOException
	 * @throws PSParseException
	 */
	static IElement getTokenFromReader(CountingPushbackReader r, IDocument d) throws IOException, PSParseException {					
		int lci;
		String v;
		if ((lci = r.read()) != -1) {
			r.unread(lci);
			char lc = (char) lci; 
			switch (lc) {
			case '<':
				v = getHexStringFromReader(r);
				break;
			case '(':
				v = getStringFromReader(r);
				break;
			case '[':
			case '{':
				lci = r.read();
				char terminator = (lc == '[') ? ']' : '}';
				arrayTerminatingCharacterStack.add(terminator);
				ArrayList<IElement> contents = new ArrayList<IElement>();
				for (;;) {
					IElement e = getTokenFromReader(r, d);
					if (e.getRepresentation().equals(String.valueOf(terminator)))
						break;
					contents.add(e);					
				}
				arrayTerminatingCharacterStack.pop();
				IElement[] ca = contents.toArray(new IElement[contents.size()]);
				if (lc == '[')
					return ElementFactory.createArrayElement(ca, d);
				else
					return ElementFactory.createProcedureElement(ca, d);								
			case '%':
				v = TerminatedStringReader.readUntilTerminator(r, commentTerminatorChecker);
				break;
			default:
				v = TerminatedStringReader.readUntilTerminator(
						r, 
						Routines.isWhitespace(lc) ?  						
							whitespaceTerminatorChecker 
							: regularTerminatorChecker
					);								
			}			
			return createElementFromString(v, d);
		}
		else {
			return null;
		}					
	}

	/**
	 * Reads one hexadecimal string
	 * 
	 * @param r source reader
	 * 
	 * @return read string
	 * 
	 * @throws IOException
	 * @throws PSParseException
	 */
	private static String getHexStringFromReader(CountingPushbackReader r) throws IOException, PSParseException {		
		if (r.eofSafeRead("Hexadecimal string wasn't started.") != '<')
			throw new IllegalStateException("Wrong start of hexadecimal string.");
		StringBuffer v = new StringBuffer();
		v.append('<');
		for (;;) {		
			char lc = r.eofSafeRead("Hexadecimal string wasn't closed.");						
			v.append(lc);
			if (lc == '>')
				break;
			if (!Routines.isLegalHexStringCharacter(lc))
				throw new PSParseException("Illegal character in hexadecimal string: " + lc);
		}
		return v.toString();
	}
	
	/**
	 * Reads one character string
	 * 
	 * @param r source reader
	 * 
	 * @return read string
	 * 
	 * @throws IOException
	 * @throws PSParseException
	 */
	private static String getStringFromReader(CountingPushbackReader r) throws IOException, PSParseException {
		int balance = 0;
		StringBuffer v = new StringBuffer();
				 
		do {
			char lc = r.eofSafeRead("balance is not null");
			
			if (lc == '\\') {
				v.append(lc);
				lc = r.eofSafeRead("empty escape sequence");
			}
			else {
				switch (lc) {
				case '(':
					balance++;
					break;
				case ')':
					balance--;
					break;
				}				
			}
			v.append(lc);
		} while (balance > 0);
		return v.toString();
	}
	
	/**
	 * Converts a character string to a normal form.
	 * 
	 * @param s PostScript representation of a string.
	 * 
	 * @return ordinary representation.
	 * 
	 * @throws PSParseException
	 */
	private static String parseCharacterString(String s) throws PSParseException {
		int[] ioc = new int[3];
		char[] oc = new char[3];		
		StringBuffer res = new StringBuffer();
		int i = 0;
		while (i < s.length()) {
			char lc = s.charAt(i++); 
			if (lc != '\\') {
				res.append(lc);				
				continue;
			}
			if (i >= s.length())
				throw new PSParseException("String terminates after '\\' character");
			lc = s.charAt(i++);
			if (!Routines.isNewLine(lc)) {
				switch (lc) {
				case 'n':
					res.append('\n');
					break;
				case 'r':
					res.append('\r');
					break;
				case 't':
					res.append('\t');
					break;
				case 'b':
					res.append('\b');
					break;
				case 'f':
					res.append('\f');
					break;
				case '\\':
					res.append('\\');
					break;
				case '0':
				case '1':
				case '2':
				case '3':
					oc[0] = lc;						
					
					int j;
					int k = 0;											
					for (j = 1; j <= 2; j++) {
						ioc[j] = s.charAt(i++);
						oc[j] = (ioc[j] == -1)? 0 : (char) ioc[j]; 						 														
						if (oc[j] < '0' || '7' < oc[j]) {
							if (ioc[j] != -1)
								i--;								
							break;
						}
						else
							k = j;	
					}
					
					int cnum = 0;
					for (j = 0; j <= k; j++)
						cnum = cnum << 3 + (oc[j] - '0');
																						
					res.append(Routines.ASCIIValues[cnum]);
					break;
					
				default:
					if (!Routines.isWhitespace(lc))
						res.append(lc);
				}			
			}
		}
		return res.toString();
	}	
}

/**
 * Maintains character table and whitespace characters list. 
 * 
 * @author Igor Sinev
 */
class Routines {

	static char[] ASCIIValues = new char[256];	
		
	static String whitespaceCharacters;
	static {
		{
			byte[] a = new byte[256];
			int i;
			for (i = 0; i < 256; i++)
				a[i] = (byte) i;
			Reader r = new InputStreamReader(new ByteArrayInputStream(a));			
			for (i = 0; i < 256; i++)
				try {
					ASCIIValues[i] = (char) r.read();
				} catch (IOException e) {
					System.out.println("Exception occured during ASCII chartable creation.");					
					e.printStackTrace();
				}
			 
		}
	
		StringBuffer b = new StringBuffer();
		Reader is = new InputStreamReader(
				new ByteArrayInputStream(
						new byte[] {0, 9, 10, 12, 13, 32}
					),
				Charset.forName("ASCII")
			);		
		try {
			int i;
			while ((i = is.read()) != -1) {
				b.append((char) i);				
			}
			whitespaceCharacters = b.toString();
		} catch (IOException e) {			
			e.printStackTrace();
		}
	}
	
	static boolean isWhitespace(char c) {
		return (whitespaceCharacters.indexOf(c) != -1);
	}
	
	static boolean isNewLine(char c) {
		return (c == 13 || c == 12 || c == 10);
	}
	
	static boolean isLegalHexStringCharacter(char c) {
		return ('0' <= c) && (c <= '9') ||
			   ('A' <= c) && (c <= 'F') ||
			   ('a' <= c) && (c <= 'f');
	}
}